obj_preps = ['до', 'к', 'об', 'на', 'про', 'в', 'во', 'с', 'по', 'относительно', 'за', 'из', 'у', 'для']
preps = ['до', 'о', 'к', 'об', 'на', 'про', 'в', 'во', 'с', 'по', 'насчет', 'относительно', 'за', 'из', 'у', 'для',
         'при', 'по', 'перед', 'после']
dets = ['какой то', 'какой-то', 'какой', 'тот', 'этот', 'это', 'такой', 'что за', 'что', 'сколько', 'чей']
modal_verbs = ['допустимый', 'возможный', 'разрешить', 'можно', 'уметь', 'мочь', 'должен', 'нужный', 'желательный',
               'надо', 'вынужденный', 'придеться', 'стоить', 'следовать', 'должен', 'необходимый', 'обязан',
               'требуется', 'предписанный', 'обязанный', 'решить', 'думать']
time_periods = ['январь', 'февраль', 'март', 'апрель', 'май', 'июнь', 'июль', 'август', 'сентябрь', 'октябрь', 'ноябрь',
                'декабрь',
                'сегодня', 'пн', 'понедельник', 'вт', 'вторник', 'ср', 'среда', 'чт', 'четверг', 'пт', 'пятница', 'сб',
                'суббота', 'вс', 'воскресенье', 'завтра']
